from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any


@dataclass
class TimeSetup:
    start: datetime
    end: datetime
    duration: timedelta

    @staticmethod
    def from_document(job_data: dict[str, Any]) -> "TimeSetup":
        start = datetime.fromisoformat(job_data["runSettings"]["startDate"])
        end = datetime.fromisoformat(job_data["runSettings"]["endDate"])
        return TimeSetup(
            start=start,
            end=end,
            duration=end - start,
        )



import logging
import sys
import tarfile
from io import BytesIO
from os import linesep
from pathlib import Path, PurePosixPath
from typing import Any, Callable, ContextManager, Optional
from urllib.parse import urlparse
from uuid import uuid4

import grpc
from colorama import Fore, Style
from colorama import init as init_colorama
from dirtools import Dir
from spm_pb2_grpc import TokenManagerStub

log = logging.getLogger("__name__")


init_colorama()


def red(msg: Any) -> str:
    return f"{Fore.RED}{msg}{Fore.RESET}"


def green(msg: Any) -> str:
    return f"{Fore.GREEN}{msg}{Fore.RESET}"


def cyan(msg: Any) -> str:
    return f"{Fore.CYAN}{msg}{Fore.RESET}"


def yellow(msg: Any) -> str:
    return f"{Fore.YELLOW}{msg}{Fore.RESET}"


def bold(msg: Any) -> str:
    return f"{Style.BRIGHT}{msg}{Style.NORMAL}"


def print_error_and_exit(msg: Any, exit_code: int = 1):
    print(red(msg))
    sys.exit(exit_code)


def print_repository_help_message(path: Path):
    print(
        f" ! {bold('Remember to change directory to the repository:')} {bold(cyan(f'cd {path}'))}"
    )
    print(f" * Connect a remote: {bold(cyan('numerous config --remote <REMOTE URL>'))}")
    print(
        f" * Checkout a scenario: {bold(cyan('numerous checkout --scenario <SCENARIO ID>'))}"
    )
    print(f" * Push changes: {bold(cyan('numerous push -c <COMMENT>'))}")
    print(f" * Build code and update scenario: {bold(cyan('numerous build'))}")


def get_grpc_channel(
    url: str, auth_plugin: Optional[grpc.AuthMetadataPlugin]
) -> grpc.Channel:
    parsed_url = urlparse(url)

    options = (
        ("grpc.keepalive_time_ms", 10000),
        ("grpc.keepalive_timeout_ms", 5000),
        ("grpc.keepalive_permit_without_calls", 1),
        ("grpc.http2_max_pings_without_data", 9999),
    )

    if parsed_url.scheme == "https":
        return grpc.secure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '443'}",
            grpc.composite_channel_credentials(
                grpc.ssl_channel_credentials(),
                grpc.metadata_call_credentials(auth_plugin),
            )
            if auth_plugin
            else grpc.ssl_channel_credentials(),
            options=options,
        )
    elif parsed_url.hostname == "localhost":
        log.debug("Getting local channel")
        return grpc.secure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '50056'}",
            credentials=grpc.local_channel_credentials(),
            options=options,
        )
    else:
        return grpc.insecure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '80'}", options=options
        )


class StubClient:
    def __init__(
        self,
        api_url: str,
        stub: Callable[[grpc.Channel], Any],
        plugin: Optional[grpc.AuthMetadataPlugin] = None,
    ):
        self.channel = get_grpc_channel(api_url, plugin)
        self.stub = stub(self.channel)

    def __enter__(self):
        return self.stub

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.channel.close()


def get_token_manager(api_url: str) -> ContextManager[TokenManagerStub]:
    return StubClient(api_url, TokenManagerStub)


class GzippedTarball:
    def __init__(self, source: Path):
        self.source = source
        self.source_dir = Dir(str(self.source))
        self.files = self.source_dir.files()
        self.tar_path = self.source / f"{uuid4()}.tar.gz"

    def __enter__(self):
        with tarfile.open(str(self.tar_path), mode="w:gz") as tar:
            for file_name in self.files:
                file_path = self.source / str(file_name)
                info = tarfile.TarInfo(str(PurePosixPath(Path(file_name))))
                info.size = file_path.stat().st_size
                with file_path.open("rb") as file:
                    tar.addfile(info, fileobj=file)
        return self.tar_path

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.tar_path.unlink()


def display_error_and_exit(
    error_header: str, help_lines: list[str], exit_code: int = 1
):
    print(red(error_header))
    print(*help_lines, sep=linesep)
    sys.exit(exit_code)


def extract_gzipped_tarball(file: BytesIO, path: Path) -> None:
    with tarfile.open(fileobj=file, mode="r:gz") as tar:
        tar.extractall(str(path))



from dataclasses import dataclass
from typing import Any


@dataclass
class SubcomponentNotFound(Exception):
    uuid: str


@dataclass
class Component:
    uuid: str
    id: str
    type: str
    name: str
    item_class: list[str]
    is_enabled: bool
    is_main: bool
    components: dict[str, "Component"]

    @staticmethod
    def from_document(
        component: dict[str, Any], all_components: list[dict[str, Any]]
    ) -> "Component":
        return Component(
            uuid=component["uuid"],
            id=component["id"],
            type=component["type"],
            name=component["name"],
            is_enabled=not bool(component["disabled"]),
            is_main=bool(component["isMainComponent"]),
            components=_extract_subcomponents(component, all_components),
            item_class=component["item_class"].split("."),
        )


def _extract_subcomponents(
    component: dict[str, Any], all_components: list[dict[str, Any]]
):
    return {
        subcomponent_ref["name"]: _extract_subcomponent(
            subcomponent_ref["uuid"], all_components
        )
        for subcomponent_ref in component["subcomponents"]
    }


def _extract_subcomponent(
    subcomponent_uuid: str, all_components: list[dict[str, Any]]
) -> Component:
    subcomponent = next(
        (
            subcomponent
            for subcomponent in all_components
            if subcomponent.get("uuid") == subcomponent_uuid
        ),
        None,
    )

    if subcomponent is None:
        raise SubcomponentNotFound(uuid=subcomponent_uuid)

    return Component.from_document(subcomponent, all_components)


def get_components_from_scenario_document(data: dict[str, Any]) -> dict[str, Component]:
    components = data.get("simComponents", [])
    subcomponent_uuids = _get_all_subcomponent_uuids(components)
    return {
        component["name"]: Component.from_document(component, components)
        for component in components
        if component["uuid"] not in subcomponent_uuids
    }


def _get_all_subcomponent_uuids(components: list[dict[str, Any]]) -> set[str]:
    all_subcomponent_uuids = set()

    for component in components:
        for subcomponent in component.get("subcomponents", []):
            all_subcomponent_uuids.add(subcomponent["uuid"])

    return all_subcomponent_uuids



import os
from dataclasses import Field, dataclass, fields


def extract_field(field: Field, value: str) -> tuple[str, int]:
    if field.name == "MAX_STATUS_MESSAGE_LENGTH":
        return (field.name, int(value))
    elif field.name == "GRPC_MAX_MESSAGE_SIZE":
        return (field.name, int(value))
    else:
        raise ValueError(f"Unexpected field {field.name}")


@dataclass
class Config:
    MAX_STATUS_MESSAGE_LENGTH: int = 32
    GRPC_MAX_MESSAGE_SIZE: int = 2**28

    @staticmethod
    def from_environment():
        extracted_fields = (
            extract_field(field, os.environ[field.name])
            for field in fields(Config)
            if field.name in os.environ
        )
        environment = {key: value for key, value in extracted_fields}
        return Config(**environment)



import sys
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import List

from dirtools import Dir, filehash
from prettytable import PrettyTable

from ..auth import AuthenticationError, login
from ..build_manager_client import HistoryUpdate, get_build_manager_client
from ..repository import NumerousRepository
from ..utils import bold, cyan, green, red, yellow


def print_updates_table(repo: NumerousRepository, updates: List[HistoryUpdate]):
    dirty = Dir(str(repo.path)).hash(index_func=filehash) != repo.snapshot
    table = PrettyTable()
    table.field_names = ["Timestamp", "Scenario", "Comment", "Commit ID", "Current"]
    for update in updates:
        table.add_row(
            [
                datetime.fromtimestamp(update.timestamp),
                update.scenario_id,
                update.comment,
                update.commit_id,
                f"{green('*')} {red('dirty') if dirty else ''}"
                if repo.commit == update.commit_id
                else "",
            ]
        )
    print(table)


def print_updates_git(repo: NumerousRepository, updates: List[HistoryUpdate]):
    dirty = Dir(str(repo.path)).hash(index_func=filehash) != repo.snapshot
    for update in updates:
        state = (
            f" {green('current')}{red(' dirty') if dirty else ''}"
            if repo.commit == update.commit_id
            else ""
        )
        print(
            f"{yellow(f'commit {update.commit_id}')} {yellow('(')}"
            f"{cyan(update.project_id + ':' + update.scenario_id)}{yellow(')')}{state}"
        )
        update_time = datetime.fromtimestamp(update.timestamp)
        print(f"Author: {update.user_full_name} <{update.user_email}>")
        print(f'Date:   {update_time.strftime("%a %b %d %H:%M:%S %Y %z")}')
        print()
        for line in update.comment.splitlines():
            print("    " + line)
        print()


def command_log(log: Logger, path: Path, display_mode: str):
    path = path or Path.cwd()

    try:
        repo = NumerousRepository(path).load()
    except Exception as excp:
        log.debug(f"Exception {type(excp)} {excp}")
        print(red(f"Cannot get logs: {bold(str(path))} is not a repository."))
        sys.exit(1)

    if repo.remote is None:
        print(red("Cannot get logs: No remote is configured for the repository."))
        print(
            f"Use the command {bold('numerous config --remote <REMOTE_URL>')} to configure a remote for the "
            "repository."
        )
        sys.exit(1)

    if repo.scenario is None:
        print(red("Cannot get logs: A scenario must be checked out."))
        print(f"Check out a scenario with {bold('numerous checkout <SCENARIO ID>')}")
        print(
            f"Push to a scenario with {bold('numerous push -c <COMMENT> <SCENARIO ID>')}"
        )
        sys.exit(1)

    try:
        id_token = login(repo)
    except AuthenticationError:
        print(red("Cannot get logs: Login failed."))
        sys.exit(1)

    with get_build_manager_client(repo.remote.api_url, id_token) as build_manager:
        updates = build_manager.get_scenario_history(repo.remote.name, repo.scenario.id)

    if not updates:
        print(f"No history exists for remote {bold(repo.remote)}")

    if display_mode == "table":
        print_updates_table(repo, updates)
    else:
        print_updates_git(repo, updates)



import logging
import threading
from collections import deque
from enum import Enum
from time import time
from typing import Callable, Iterator, Optional
from urllib.parse import urlparse

import grpc
from numerous.client_common.validation_interceptor import ValidationInterceptor

from . import config
from .config import GRPC_MAX_MESSAGE_SIZE

log = logging.getLogger("numerous.client")


class Interp(Enum):
    zero = 0
    linear = 1
    quadratic = 2


class JobStatus(Enum):
    ready = 0
    running = 1
    finished = 2
    request_termination = 3
    terminated = 4
    failed = 5
    requested = 6
    initializing = 7


class RepeatedFunction:
    def __init__(self, interval, function, run_initially=False, *args, **kwargs):
        self._timer: Optional[threading.Timer] = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time()

        if run_initially:
            self.function(*self.args, **self.kwargs)

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        if self._timer is not None:
            self._timer.cancel()
        self.is_running = False


class RestorableIterator(Iterator):
    """
    Returns items not handled items again after restore.
    >>> restorable_iterator = RestorableIterator(iter(range(4)))
    >>> assert list(restorable_iterator) == [0, 1, 2, 3]
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [0, 1, 2, 3]
    >>> restorable_iterator.restore()
    >>> assert list(next(restorable_iterator) for i in range(3)) == [0, 1, 2]
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [2, 3]
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [3]
    >>> assert list(restorable_iterator) == []
    """

    def __init__(self, data: Iterator):
        self._not_handled: deque = deque()
        self._restored_data: deque = deque()
        self._data = data
        self._lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self._lock:
            next_item = (
                self._restored_data.popleft()
                if self._restored_data
                else next(self._data)
            )
            self._not_handled.append(next_item)
            return next_item

    def item_handled(self):
        with self._lock:
            self._not_handled.popleft()

    def restore(self):
        with self._lock:
            self._restored_data = self._not_handled + self._restored_data
            self._not_handled = deque()


def initialize_grpc_channel(
    refresh_token: str,
    token_callback: Callable[[], str],
    server: Optional[str] = None,
    port: Optional[int] = None,
    secure: Optional[bool] = None,
    instance_id: Optional[str] = None,
):
    secure_channel = secure if secure is not None else not config.FORCE_INSECURE
    if secure_channel is None:
        raise RuntimeError("Secure channel must be specified")
    server = server or config.NUMEROUS_API_SERVER
    if server is None:
        raise RuntimeError("Server must be specified")
    port = port or config.NUMEROUS_API_PORT
    if port is None:
        raise RuntimeError("Port must be specified")

    log.debug("Initializing gRPC channel %s:%s", server, port)

    options = [
        ("grpc.max_message_length", GRPC_MAX_MESSAGE_SIZE),
        ("grpc.max_send_message_length", GRPC_MAX_MESSAGE_SIZE),
        ("grpc.max_receive_message_length", GRPC_MAX_MESSAGE_SIZE),
    ]

    if secure_channel:
        creds = grpc.ssl_channel_credentials()
        channel = grpc.secure_channel(f"{server}:{port}", creds, options)
    else:
        channel = grpc.insecure_channel(f"{server}:{port}", options)

    vi = ValidationInterceptor(
        token=refresh_token, token_callback=token_callback, instance=instance_id
    )
    channel = grpc.intercept_channel(channel, vi)
    return channel, vi.instance


def parse_api_url(url: str):
    parsed_url = urlparse(url)
    server = parsed_url.hostname
    port = parsed_url.port or (443 if parsed_url.scheme == "https" else 80)
    secure = parsed_url.scheme == "https"
    return server, port, secure



import json
import sys
from datetime import datetime
from logging import Logger, getLogger
from pathlib import Path
from typing import Optional

import requests
from dirtools import Dir, DirState, filehash

from ..auth import AuthenticationError, login
from ..build_manager_client import (
    SnapshotAlreadyExistsError,
    SnapshotError,
    get_build_manager_client,
)
from ..repository import NumerousRepository
from ..utils import GzippedTarball, bold, cyan, green, red, yellow

log = getLogger(__name__)


def get_snapshot_id(path: Path) -> str:
    dir = Dir(str(path))
    return str(dir.hash(index_func=filehash))


def get_snapshot_state(path: Path) -> DirState:
    dir = Dir(str(path))
    state = DirState(dir, index_cmp=filehash).state.copy()
    del state["directory"]
    return state


def print_push_error(msg: str, help_text: Optional[str] = None) -> None:
    print(red(f"Cannot push: {msg}"))
    if help_text:
        print(help_text)


def upload_snapshot(path: Path, upload_url: str, snapshot_id: str) -> None:
    print("Uploading snapshot...")
    with GzippedTarball(path) as tar_path:
        log.debug(f"Created tarball {tar_path} for {path}")
        with open(tar_path, "rb") as tar_file:
            upload_response = requests.post(
                upload_url,
                data=tar_file,
                headers={"Content-Type": "multipart/related"},
                params={"name": snapshot_id, "mimeType": "application/octet-stream"},
            )
            if upload_response.ok:
                log.debug(
                    f"Uploaded snapshot {snapshot_id} got response: {upload_response.status_code} "
                    f"{upload_response.reason}"
                )
            else:
                log.debug(
                    f"Failed uploading {snapshot_id} {upload_response.status_code} {upload_response.reason}"
                )
                raise RuntimeError("An error occured during upload")


def command_push(log: Logger, repo_path: Optional[Path], comment: str) -> None:
    path = Path.cwd() if repo_path is None else repo_path

    if not path.exists():
        print_push_error(f"{bold(path)} does not exist")
        sys.exit(1)

    repo = NumerousRepository(path).load()

    if repo.scenario is None:
        print_push_error("A scenario must be checked out.")
        sys.exit(1)

    if repo.snapshot == Dir(str(path)).hash(index_func=filehash):
        print_push_error("Cannot push: There is nothing new to push.")
        sys.exit(1)

    if repo.remote is None:
        help_text = (
            f'Use the command {bold("numerous config --remote <REMOTE_URL>")} to configure a remote for the '
            "repository."
        )
        print_push_error("No remote is configured for the repository.", help_text)
        sys.exit(1)

    try:
        id_token = login(repo)
    except AuthenticationError:
        print_push_error("Login failed.")
        sys.exit(1)

    print(cyan(f"Pushing to remote repository {bold(repo.remote)}..."))
    print("Creating snapshot metadata...")
    snapshot_id = get_snapshot_id(repo.path)
    snapshot_state = json.dumps(get_snapshot_state(repo.path))

    if len(snapshot_state) > 4e6:
        print_push_error(
            "The snapshot contains too many files and directories - please use the .exclude file to "
            "exclude files that are not needed."
        )
        sys.exit(1)

    with get_build_manager_client(repo.remote.api_url, id_token) as build_manager:
        try:
            upload_url = build_manager.create_snapshot_get_upload_url(
                repo.remote.name, snapshot_id, snapshot_state
            )
        except SnapshotAlreadyExistsError:
            print(yellow(f"Snapshot {snapshot_id} already exists. Stopping."))
            sys.exit(0)
        except SnapshotError:
            print_push_error("An error occured during snapshot creation.")
            sys.exit(1)

        try:
            upload_snapshot(repo.path, upload_url, snapshot_id)
        except Exception as e:
            print_push_error(str(e))
            sys.exit(1)

        commit_id = build_manager.create_history_update(
            repo.remote.name,
            repo.scenario.project_id,
            repo.scenario.id,
            snapshot_id,
            repo.commit,
            comment,
            datetime.utcnow().timestamp(),
        )
    print(green(f"Pushed snapshot {bold(snapshot_id)} as commit {bold(commit_id)}."))

    repo.commit = commit_id
    repo.snapshot = snapshot_id
    repo.save()



from .example_models import (  # TODO: fix missing/previously overridden import of .example_models.multiply_by_two
    model,
)
from .examples import make_report, multiply_by_two
from .report import report

__all__ = ("model", "multiply_by_two", "make_report", "report")



from dataclasses import dataclass
from typing import Any

from .component import Component, get_components_from_scenario_document


@dataclass
class Scenario:
    id: str
    name: str
    components: dict[str, Component]

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Scenario":
        return Scenario(
            id=data["id"],
            name=data.get("scenarioName", ""),
            components=get_components_from_scenario_document(data),
        )
